{{ SumoMouse.spin }}

'' This object provides a low level assembly driver to control a ADNS-2610 mouse
'' sensor. Mouse position is integrated over time and rate of change is available
'' with an adjustable sample period. Several imaging sensor statistics are also
'' available: squall, minimum pixel, maximum pixel, average pixel (actually it is
'' the sum of pixel values. The true average pixel = sum * 256 / 324)
''
'' The assembly driver retrieves the current sensor values from the mouse sensor
'' and updates the shared memory buffer. The counter value is updated every time
'' the sensor data is updated if synchronization is required. The LSB of the counter
'' is incremented every time the rate value is update.  

CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 6_000_000

VAR
  long  cog
  
PUB start(buff_ptr) : okay

'' Start driver - starts a cog
'' returns false if no cog available
'' may be called again to change settings
''

  stop
  return cog := cognew(@entry, buff_ptr) + 1


PUB stop

'' Stop driver - frees a cog

  if cog
    cogstop(cog~ - 1)


DAT

'*********************************************************
'* Assembly language driver for ADNS-2610 mouse sensor   *
'*********************************************************

                        org
'
'
' Read initialization parameters
'
entry                  

' Setup I/O pin directions
                        or      outa, clk_pin
                        or      outa, dout_pin

                        andn    dira, din_pin           
                        andn    dira, dout_pin
                        or      dira, clk_pin

                        mov     frame_ptr, par
                        add     frame_ptr, #4 * 8


' Reset the mouse sensor
                        mov     t1, cnt
                        add     t1, t_sptt
                        waitcnt t1, #0

                        call    #reset

                        mov     t1, cnt
                        add     t1, t_sptt
                        waitcnt t1, #0
                        
       
' Setup CTRA to time sample rate
                        mov     ctra,       CounterMode     ' Setup counter to time pwm cycles
                        mov     frqa, #1
                        mov     phsa, #0
                        
main_loop
                        'Change Mouse Status
                        mov     p1, par
                        add     p1, #4 * 7
                        rdlong  t1, p1
                        cmp     lastmode, t1 wc,wz
              if_ne     mov     lastmode, t1
              if_ne     mov     cmd, #0
              if_ne     mov     bits, t1
              if_ne     call    #write_cmd
              
                        'Wait for serial port reset timeout

                        call    #read_xy
                        call    #read_status

                        mov     p1, par
                        add     p1, #4 * 1
                        rdlong  sample_period,  p1
                        
                        cmp     sample_period,  phsa wc, wz         ' Check for servo period reset
              if_be     mov     phsa, #0                              
              if_be     call    #update_rate

                        mov     p1, par                 ' Update Mouse Data Counter
                        rdlong  t1, p1
                        add     t1, #2                  ' Increment counter by 2 every 
              if_be     add     t1, #1                  ' Toggle last bit on rate update 
                        wrlong  t1, p1
                        
                        mov     t1,cnt
                        add     t1, t_hold
                        waitcnt t1, #0
                                                                                               
                        jmp     #main_loop                        

update_rate
                        mov     p1, par
                        add     p1, #4*4
                        wrlong  x_rate, p1

                        add     p1, #4
                        wrlong  y_rate, p1

                        'call    #update_theta

                        {{
                        abs     t1, x_rate
                        cmp     rate_limit, t1 wc, wz
              if_be     call    #reset
              if_be     xor     dira, led_pin

                        abs     t1, y_rate
                        cmps    rate_limit, t1 wc, wz
              if_be     call    #reset
              if_be     xor     dira, led_pin     
                        }}
                                                  
                                                
                        mov     x_rate, #0
                        mov     y_rate, #0                         
update_rate_ret         ret


update_theta
                        mov     y, x_rate       ' theta_rate = x_rate * k_theta / 1024 
                        cmps    y,  #0   wc, wz
                        mov     t3, #0
              if_b      mov     t3, #1
              if_b      abs     y, y
                            
                        mov     x, k_theta
                        call    #multiply                         
                        shr     y, #10          ' Divide by 1024

                        test    t3, #1 wz
              if_nz     neg     y, y

                        'mov     t1, par
                        'add     t1, #4 * 4
                        'rdlong  theta, t1
                        'adds    theta, y
                        'wrlong  theta, t1
update_theta_ret        ret


reset
                        mov     cmd, #0
                        mov     bits, #%1_0_00000_0
                        call    #write_cmd
reset_ret               ret

awake
                        mov     cmd, #0
                        mov     bits, #%0_0_00000_1
                        call    #write_cmd
awake_ret               ret

shutdown                mov     cmd, #0
                        mov     bits, #%0_1_00000_0
                        call    #write_cmd
shutdown_ret            ret

startup                 mov     cmd, #0
                        mov     bits, #%0_0_00000_0
                        call    #write_cmd
startup_ret             ret

read_status
                        mov     p1, par
                        add     p1, #4*6
                        
                      ' Squall
                        mov     cmd, #04
                        call    #read_cmd
                        wrbyte  bits, p1
                        add     p1, #1

                        ' Minimum Pixel
                        mov     cmd, #06
                        call    #read_cmd
                        wrbyte  bits, p1
                        add     p1, #1

                        ' Avgerage Pixel
                        mov     cmd, #07
                        call    #read_cmd
                        wrbyte  bits, p1
                        add     p1, #1
                        
                      ' Maximum Pixel
                        mov     cmd, #05
                        call    #read_cmd
                        wrbyte  bits, p1
                        add     p1, #1



                        
                                                 
read_status_ret         ret

read_xy
                        ' X Movement
                        mov     cmd, #02
                        call    #read_cmd
                        test    bits, #80 wz
              if_nz     or      bits, maskneg                        
                        
                        mov     t1, par
                        add     t1, #4 * 2
                        
                        rdlong  t2, t1          'Update absolute position
                        adds    t2, bits
                        wrlong  t2, t1
                        
                        adds    x_rate, bits    'Update rate counter

                                
                        ' Y Movement
                        mov     cmd, #03
                        call    #read_cmd
                        test    bits, #80 wz
              if_nz     or      bits, maskneg                        

                        mov     t1, par
                        add     t1, #4 * 3

                        rdlong  t2, t1          'Update absolute position
                        adds    t2, bits
                        wrlong  t2, t1
                        adds    y_rate, bits    'Update rate counter
                        
read_xy_ret             ret


read_pixels

                        mov     cmd, #08                ' Read Pixel Data
                        call    #read_cmd

                        test    bits, #%0100_0000 wz    ' Check if the data byte is valid
              if_z      jmp     #:end_loop             

                        test    bits, #%1000_0000 wz    ' Check if the start of frame bit is set
              if_z      jmp     #:next_byte             ' otherwise read the next byte
              
                        test    frame_mode, #01 wz            ' if mode is set then grab a new frame
              if_nz     jmp     #:grab_frame
                  
                        mov     frame_mode, #01               ' this is the first valid pixel after a new frame
                        jmp     #:next_byte
                                                
:grab_frame             call    #grab_frame

                        mov     frame_mode, #00
                        mov     frame_ptr, par          ' reset frame buffer pointer
                        add     frame_ptr, #4 * 6
                                                
                        jmp     #:end_loop
                        
                        
:next_byte              and     bits, #%0011_1111
                        wrbyte  bits, frame_ptr
                        add     frame_ptr, #1
                        
                        jmp     #:end_loop
                        
                               
:end_loop


read_pixels_ret         ret

grab_frame
                        mov     cmd,  #08
                        mov     bits, #00
                        call    #write_cmd
grab_frame_ret          ret


write_cmd               mov     t2, cmd
                        or      t2, #%1000_0000
                        call    #write
                        mov     t2, bits
                        call    #write
                        mov     t1, cnt
                        add     t1, t_sw
                        waitcnt t1, #0
write_cmd_ret           ret
                                                
read_cmd
                        mov     t2, cmd
                        call    #write
                        call    #read
                        mov     bits, t2                        
read_cmd_ret            ret

read
                        mov     t1, #8
                        mov     t2, #0

                        mov     t3, cnt                 ' Delay for read processing
                        add     t3, t_hold
                        waitcnt t3, t_clk

                        
:loop                   andn    outa, clk_pin
                        waitcnt t3, t_clk

                        or      outa, clk_pin
                        
                        shl     t2, #1
                        mov     t4, ina
                        test    t4, din_pin wz
                        muxnz   t2, #1

                        waitcnt t3, t_clk
                        djnz    t1, #:loop

           
                        
                        
read_ret                ret

write
                        andn    outa, clk_pin

                        mov     t3, cnt                 ' Delay for SDIO handoff
                        add     t3, t_delay
                        waitcnt t3, t_clk

                        or      dira, dout_pin          ' Make data pin output
                                                
                        mov     t1, #8
:loop                   test    t2, #$80 wz             ' Shift data high bit first
                        muxnz   outa, dout_pin
                        shl     t2, #1

                        andn    outa, clk_pin
                        waitcnt t3, t_clk
                        
                        or      outa, clk_pin
                        waitcnt t3, t_clk
                                                              
                        djnz    t1, #:loop

                        andn    dira, dout_pin          ' Make data pin input                    
write_ret               ret

'
' Get a parameter
'
param                   rdlong  t3,t1                   'get parameter into t3
                        add     t1,#4                   'point to next parameter
param_ret               ret

{ Multiply x[15..0] by y[15..0] y[31..16] must be 0
  on exit, product in y[31..0] }
multiply      shl       x,          #16
              mov       t1,          #16
              shr       y,          #1  wc
:loop   if_c  add       y,          x   wc
              rcr       y,          #1  wc
              djnz      t1,          #:loop 
multiply_ret  ret

'
'
' Initialized data
'
CounterMode   long      %0_11111_000_00000000_000000_000_000000

dlsb                    long    1 << 9
mask12                  long    $FFF

lastmode                long    0
t_hold                  long    96 * 100
t_delay                 long    96 * 10
t_sw                    long    96 * 100
t_long                  long    96 * 1000

t_sptt                  long    96_000 * 90
t_clk                   long    96 * 10

sample_period           long    96 * 1000 * 10

clk_pin                 long    1 << 16
din_pin                 long    1 << 17   ' 17 on Proto V1
dout_pin                long    1 << 17   ' 18 on Proto V1

'led_pin                long    1 << 31

maskneg                 long    $FFFFFF00

x_rate                  long    0
y_rate                  long    0
theta                   long    0
radius                  long    0

k_theta                 long    740       ' calibration to convert x_rate to 1/10 degree


rate_limit              long    40

'
'
' Uninitialized data
'
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1

p1                      res     1

frame_ptr               res     1
frame_mode              res     1

cmd                     res     1
bits                    res     1
x                       res     1
y                       res     1