{{ SonarDriver.spin }}
CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 6_000_000


VAR
  long cog                     'cog flag/id

    
PUB start(buff_ptr) : okay
'' Start the sonar driver - starts a cog
'' returns false if no cog available
  stop
  okay := cog := cognew(@entry, buff_ptr) + 1

PUB stop

'' Stop serial driver - frees a cog

  if cog
    cogstop(cog~ - 1)

DAT

entry         org       0
              
{ Initialization }
  '           or        dira,       Led             'Use Led pin for debugging and timing
                            
              mov       p0,         par
              rdlong    dist_buff,  p0              'Pointer to Speed Buffer
              add       p0,         #4

   
              andn      outa,       s0_out          'Clear all outputs 
              
              or        dira,       s0_out          'Set I/O Pin directions
              andn      dira,       s0_in         
   
              mov       s_out,      s0_out          ' Create valid sonar output pin mask
   
              mov       ctra,       CounterMode     ' Setup counter to time pulses
              mov       frqa,       #1
              mov       window,     cnt             ' Setup intial window
              add       window,     #9

{ Start Main Loop }
MainLoop
              waitcnt   window,     Period          ' Wait for start of window (20 Hz)

{ Check which sonars are connected }

CheckSonar
              mov       s_in,       s0_in           ' default input pin masks for all sonars valid
              
              mov       s0_val,     Timeout         ' sentinel value for sonar timeout
            
              mov       t0,         ina
              and       t0,         s0_in wz, nr    ' if bit is high then sonar isn't connected
        if_nz andn      s_in,       s0_in           '
        if_nz mov       s0_val,     #0              ' sentinel value for failed sonar

{ Get a pulse width from all connected sonars by triggering all the sonars and
  then timing the sonar output pulse using ctra. The accuracy of measuring pulse
  widths is limited by the average loop time of 1uSec or 1/100th inch }

PingSonar
              or        outa,       s_out           ' Trigger all sonars simultaneously
              mov       c0,         cnt
              add       c0,         Pulse             
              waitcnt   c0,         Blank
              andn      outa,       s_out

              waitpeq   s_in,       s_in            ' Wait for start of sonar timing pulses
              mov       phsa,       #0

              
:sonar0       test      s_in,       s0_in wz        ' Check if still monitoring this sonar
        if_z  jmp       #:done

              mov       t0,         ina             ' Check if the timing pulse is still high
              and       t0,         s0_in wz

        if_z  mov       s0_val,     phsa            ' Save counter value when pulse goes low
        if_z  andn      s_in,       s0_in           ' Stop watching this sonar
 
:done
              mov       c0,         phsa            ' Check if loop has exceeded maximum timeout
              cmp       c0,         Timeout  wz,wc  ' This prevents a lockout situation on failure
        if_a  jmp       #ConvertToInches

              tjnz      s_in,       #:sonar0        ' Loop until all valid sonars have returned
                                                    ' a value or have timedout

{ Convert the pulse width from clock cycles (96Mhz) to 1/100th inches
  The time to distance conversion is 1 in = 148 uSec }

ConvertToInches                                     ' Requires 22uSec to convert to 1/100th inches
              mov       x,          s0_val
              call      #convert
              mov       s0_val,     x


{ Update the global memory buffer with the latest sonar readings }
              
UpdateDist
              mov       p0,         par              ' Read Ping Frequency
              add       p0,         #4
              rdlong    Period,     p0

              add       p0,         #4               ' Update Sonar Delta
              mov       t1,         s0_val
              subs      t1,         s0_last
              wrlong    t1,         p0
              mov       s0_last,    s0_val
              
              add       p0,         #4               ' Update Sonar Value
              wrword    s0_val,     p0

              add       p0,         #2               ' Read Average Limit
              rdword    t1,         p0

              cmp       s0_cnt,     t1 wz,wc
       if_b   jmp       #:skip_write

              mov       x,          s0_sum           ' Calculate Average
              mov       y,          s0_cnt
              call      #divide
              mov       s0_sum,     x
              
              add       p0,         #2               ' Update sonar average value
              wrword    s0_sum,     p0
              mov       s0_cnt,     #0
              mov       s0_sum,     #0
:skip_write
              add       s0_cnt,     #1
              add       s0_sum,     s0_val
              
              mov       p0,         par              ' Update Change Counter
              rdlong    t1,         p0
              add       t1,         #1
              wrlong    t1,         p0

{ Check that the timing window hasn't been exceeded. Otherwise the cog will be locked
  until the counter rolls over }
  
CheckWindow
              mov       c0,         cnt
              add       c0,         #20
              min       window, c0
              
              jmp       #MainLoop            

{ Subroutines }

{ Convert from clock cycles to inches.
  In:  x - pulse width in clock cycles (96 Mhz)
  Out: x - distance in 1/100th inch }
convert
              mov       y,          UnitD           ' val / 14,208
              call      #divide
              mov       t0,         x               ' save quotient results
              and       t0,         LoWord            
              shr       x,          #16             ' convert remainder to fixed precision
              mov       y,          FracD           ' remainder / 142
              call      #divide
              mov       t1,         x               ' save quotient results
              and       t1,         LoWord          ' discard remainder
              mov       x,          t0              ' convert quotient to fixed precision
              mov       y,          UnitN
              call      #multiply
              mov       x,          y
              add       x,          t1
              
convert_ret   ret
              
{ Multiply x[15..0] by y[15..0] y[31..16] must be 0
  on exit, product in y[31..0] }
multiply      shl       x,          #16
              mov       t,          #16
              shr       y,          #1  wc
:loop   if_c  add       y,          x   wc
              rcr       y,          #1  wc
              djnz      t,          #:loop 
multiply_ret  ret

{ Divide x[31..0] by y[15..0] (y[16] must be 0)
  on exit, quotient is in x[15..0] and remainder is in
  x[31..16] }

divide        shl       y,          #15
              mov       t,          #16
:loop         cmpsub    x,          y wc
              rcl       x,          #1
              djnz      t,          #:loop
divide_ret    ret

'Initialized Variables
Period        long      96*50_000-9  '50 mSec - 20 Hz
Timeout       long      96*30_000    '30 mSec
CounterMode   long      %0_11111_000_00000000_000000_000_000000
Pulse         long      96*20-9     '20uSec pulse
Blank         long      96*100-9    '100 uSec blanking period
UnitN         long      100
UnitD         long      14_208      '96*148 = 14_208
FracD         long      142         '100 / 14_208
LoWord        long      $FFFF

s0_in         long      1 << 25     'pin mask for sonar io
s0_out        long      1 << 9      'pin mask for sonar io


sentinel      long      270         'Sentinel Value for timeout

'Temporary Variables
p0            res       1

t0            res       1
t1            res       1
t2            res       1

c0            res       1
delta         res       1

x             res       1
y             res       1
t             res       1

'State Variables
window        res       1           'holds counter value of next window
dist_buff     res       1           'pointer to output buffer (sonars+1 longs)

s_in          res       1
s_out         res       1


s0_val        res       1           'holds last pulse measurement
s0_sum        res       1
s0_cnt        res       1
s0_last       res       1